/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "globus_gridftp_server_s3_base.h"

typedef
struct globus_l_s3_multipart_s
{
    int64_t                             initiated_time;
    struct globus_l_s3_multipart_s
                                       *next;
    char                                upload_id[];
}
globus_l_s3_multipart_t;

typedef
struct
{
    char                               *pathname;
    char                               *key;
    char                               *next_key_marker;
    char                               *next_upload_id_marker;
    int                                 retries;
    S3Status                            status;
    globus_result_t                     result;
    struct globus_l_s3_multipart_s
                                       *partial_uploads;
}
globus_l_s3_list_multipart_t;

static
S3Status
globus_l_s3_recv_multipart_list_uploads_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data);

static
void
globus_l_s3_recv_multipart_list_uploads_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

static
S3Status
globus_l_s3_recv_multipart_list_uploads_data(
    int                                 is_truncated,
    const char                         *next_key_marker,
    const char                         *next_upload_id_marker,
    int                                 uploads_count, 
    const S3ListMultipartUpload        *uploads,
    int                                 common_prefixes_count,
    const char                        **common_prefixes,
    void                               *callback_data);

static
S3Status
globus_l_s3_recv_multipart_abort_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data);

static
void
globus_l_s3_recv_multipart_abort_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data);

globus_result_t
globus_i_s3_recv_multipart_resolve_id(
    globus_i_s3_transfer_t             *transfer_state)
{
    globus_l_s3_list_multipart_t        list_multipart_state;
    globus_l_s3_multipart_t            *oldest_upload = NULL;

    list_multipart_state = (globus_l_s3_list_multipart_t)
    {
        .pathname                     = transfer_state->transfer_info->pathname,
        .key                          = transfer_state->key
    };

    do
    {
        char                           *next_key_marker = list_multipart_state.next_key_marker;
        char                           *next_id_marker = list_multipart_state.next_upload_id_marker;
        list_multipart_state.next_key_marker = NULL;
        list_multipart_state.next_upload_id_marker = NULL;
        
        list_multipart_state.retries = 0;

        do
        {
            list_multipart_state.status = S3StatusOK;
            list_multipart_state.result = GLOBUS_SUCCESS;

            globus_i_s3_retry_delay(list_multipart_state.retries);

            GlobusS3Info("REST: op=%p S3_list_multipart_uploads(bucket=%s, key=%s) try=%d\n",
                        (void *) transfer_state->op,
                        transfer_state->bucket_name,
                        transfer_state->key,
                        transfer_state->retries);

            S3_list_multipart_uploads(
                    &transfer_state->bucket_context,
                    transfer_state->key,
                    next_key_marker,
                    next_id_marker,
                    NULL,
                    "/",
                    0,
                    NULL,
                    &(S3ListMultipartUploadsHandler)
                    {
                        .responseHandler =
                        {
                         .propertiesCallback
                                      = globus_l_s3_recv_multipart_list_uploads_properties,
                        .completeCallback
                                      = globus_l_s3_recv_multipart_list_uploads_complete
                        },
                    .responseXmlCallback
                                      = globus_l_s3_recv_multipart_list_uploads_data
                    },
                    &list_multipart_state);

            GlobusS3DebugOpStatus(S3_list_multipart_uploads, transfer_state);
        }
        while (S3_status_is_retryable(list_multipart_state.status)
                && (list_multipart_state.retries++) < globus_i_s3_max_retries);

        free(next_key_marker);
        free(next_id_marker);

        while (list_multipart_state.partial_uploads != NULL)
        {
            globus_l_s3_multipart_t    *newer_upload = NULL;

            if (oldest_upload == NULL
                || oldest_upload->initiated_time > list_multipart_state.partial_uploads->initiated_time)
            {
                newer_upload = oldest_upload;
                oldest_upload = list_multipart_state.partial_uploads;
                list_multipart_state.partial_uploads = oldest_upload->next;
                oldest_upload->next = NULL;
            }
            else
            {
                newer_upload = list_multipart_state.partial_uploads;
                list_multipart_state.partial_uploads = list_multipart_state.partial_uploads->next;
            }

            if (newer_upload)
            {
                GlobusS3Info("REST: op=%p S3_abort_multipart_upload(id=%s) try=%d\n",
                            (void *) transfer_state->op,
                            transfer_state->multipart_id,
                            0);

                S3_abort_multipart_upload(
                    &transfer_state->bucket_context,
                    transfer_state->key,
                    newer_upload->upload_id,
                    &(S3AbortMultipartUploadHandler)
                    {
                        .responseHandler = 
                        {
                            .propertiesCallback
                                      = globus_l_s3_recv_multipart_abort_properties,
                            .completeCallback
                                      = globus_l_s3_recv_multipart_abort_complete
                        }
                    });
                free(newer_upload);
            }
        }
    }
    while (list_multipart_state.result == GLOBUS_SUCCESS
        && (list_multipart_state.next_key_marker != NULL
            || list_multipart_state.next_upload_id_marker != NULL));

    
    if (oldest_upload != NULL)
    {
        transfer_state->multipart_id = strdup(oldest_upload->upload_id);
        if (transfer_state->multipart_id == NULL)
        {
            return GlobusGFSErrorMemory("multipart_id");
        }
        free(oldest_upload);
    }
    else
    {
        return GlobusGFSErrorGeneric("No multi-part upload found for restart");
    }
    return GLOBUS_SUCCESS;
}
/* globus_i_s3_recv_multipart_resolve_id() */

static
S3Status
globus_l_s3_recv_multipart_list_uploads_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_list_uploads_properties() */

static
void
globus_l_s3_recv_multipart_list_uploads_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_l_s3_list_multipart_t       *list_state = callback_data;

    GlobusS3Enter();

    assert(list_state != NULL);

    if (status != S3StatusOK && list_state->result == GLOBUS_SUCCESS)
    {
        list_state->status = status;
        list_state->result = GlobusGFSErrorS3(status, list_state->pathname, error_details);
    }

    GlobusS3Exit();

    return;
}
/* globus_l_s3_recv_multipart_list_uploads_complete() */

static
S3Status
globus_l_s3_recv_multipart_list_uploads_data(
    int                                 is_truncated,
    const char                         *next_key_marker,
    const char                         *next_upload_id_marker,
    int                                 uploads_count, 
    const S3ListMultipartUpload        *uploads,
    int                                 common_prefixes_count,
    const char                        **common_prefixes,
    void                               *callback_data)
{
    globus_l_s3_list_multipart_t       *list_state = callback_data;
    globus_l_s3_multipart_t            *list_entry = NULL;
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    assert(list_state != NULL);

    for (int i = 0; i < uploads_count; i++)
    {
        if (strcmp(uploads[i].key, list_state->key) == 0)
        {
            list_entry = malloc(sizeof(globus_l_s3_multipart_t) + strlen(uploads[i].uploadId) + 1);

            if (list_entry == NULL)
            {
                list_state->result = GlobusGFSErrorMemory("list_entry");
                status = S3StatusOutOfMemory;

                goto done;
            }
            *list_entry = (globus_l_s3_multipart_t)
            {
                .initiated_time = uploads[i].initiated,
                .next = list_state->partial_uploads
            };
            strcpy(list_entry->upload_id, uploads[i].uploadId);
            list_state->partial_uploads = list_entry;
            list_entry = NULL;
        }
    }
    if (is_truncated)
    {
        if (next_key_marker != NULL)
        {
            list_state->next_key_marker = strdup(next_key_marker);
            if (list_state->next_key_marker == NULL)
            {
                status = S3StatusOutOfMemory;
                list_state->result = GlobusGFSErrorMemory("next_key_marker");
                goto done;
            }
        }
        if (next_upload_id_marker != NULL)
        {
            list_state->next_upload_id_marker = strdup(next_upload_id_marker);
            if (list_state->next_upload_id_marker == NULL)
            {
                status = S3StatusOutOfMemory;
                list_state->result = GlobusGFSErrorMemory("next_upload_id_marker");
                goto done;
            }
        }
    }

done:
    GlobusS3ExitStatus(status);
    return status;
}
/* globus_l_s3_recv_multipart_list_uploads_data() */

static
S3Status
globus_l_s3_recv_multipart_abort_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_data)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_recv_multipart_abort_properties() */

static
void
globus_l_s3_recv_multipart_abort_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    GlobusS3Enter();

    GlobusS3Exit();

    return;
}
/* globus_l_s3_recv_multipart_abort_complete() */
// vim: ft=c:
