/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @file rmdir.c
 * Implementation of rmdir for the S3 DSI
 *
 * For non-bucket rmdirs, this will simply delete the placeholder file if it
 * exists. There may remain objects that contain the prefix of the "directory"
 * that we deleted.
 *
 * We do not error if the placeholder does not exist.
 */
#include "globus_gridftp_server_s3_base.h"

static
void *
globus_l_s3_rmdir_thread(
    void                               *user_arg);

static
void
globus_l_s3_rmdir_complete(
    S3Status                            status,
    const S3ErrorDetails               *errorDetails,
    void                               *callbackData);

static
S3Status
globus_l_s3_rmdir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg);

typedef struct
{
    /** GridFTP rmdir operation */
    globus_gfs_operation_t              op;
    /** Bucket Context derived from the session info and the rmdir pathname */
    S3BucketContext                     bucket_context;
    /** pointer into pathname that contains the bucket name */
    char                               *bucket_name;
    /** pointer into pathname that contains the key name */
    char                               *key;
    /** Number of retries for current operation */
    int                                 retries;
    /** S3 operation status */
    S3Status                            status;
    /** Result of the operation */
    globus_result_t                     result;
    /** Copy of the pathname string split into bucket_name and key */
    char                                pathname[];
}
globus_l_s3_rmdir_t;
/**
 * @brief rmdir
 * @details
 * For s3, there are not really any directories, so we just try to see if
 * the prefix pathname/ contains anything, and if not, remove pathname/ it
 */
void
globus_i_s3_rmdir(
    globus_gfs_operation_t              op,
    globus_gfs_command_info_t          *cmd_info,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_l_s3_rmdir_t                *rmdir_state = NULL;
    globus_thread_t                     t;
    int                                 rc = 0;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    GlobusS3Info("rmdir op=%p pathname=%s\n",
                (void *) op,
                cmd_info->pathname);

    assert(s3_handle != NULL);

    rmdir_state = malloc(sizeof(globus_l_s3_rmdir_t) + strlen(cmd_info->pathname) + 2);
    if (rmdir_state == NULL)
    {
        result = GlobusGFSErrorMemory("rmdir_state");

        goto failure;
    }

    *rmdir_state = (globus_l_s3_rmdir_t)
    {
        .op                           = op
    };

    strcpy(rmdir_state->pathname, cmd_info->pathname);

    globus_i_s3_path_parse(rmdir_state->pathname, &rmdir_state->bucket_name, &rmdir_state->key);

    rmdir_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, rmdir_state->bucket_name);

    if (rmdir_state->key)
    {
        strcat(rmdir_state->key, "/");
    }

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_rmdir_thread,
            rmdir_state);

    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorGeneric("globus_thread_create");
failure:
        
        GlobusS3Info("finished_command: op=%p result=%#x\n",
                    (void *) op,
                    (unsigned int) result);

        globus_gridftp_server_finished_command( 
                op,
                result,
                NULL);
    }

    GlobusS3Exit();

    return;
}
/* globus_i_s3_rmdir() */

static
void *
globus_l_s3_rmdir_thread(
    void                               *user_arg)
{
    globus_l_s3_rmdir_t                *rmdir_state = user_arg;

    if (rmdir_state->key)
    {
        rmdir_state->retries = 0;
        do
        {
            rmdir_state->status = S3StatusOK;
            rmdir_state->result = GLOBUS_SUCCESS;
            
            globus_i_s3_retry_delay(rmdir_state->retries);

            GlobusS3Info("REST: op=%p S3_delete_object(bucket=%s, key=%s) try=%d\n",
                         (void *) rmdir_state->op,
                         rmdir_state->bucket_name,
                         rmdir_state->key,
                         rmdir_state->retries);
            S3_delete_object(
                    &rmdir_state->bucket_context,
                    rmdir_state->key,
                    NULL,
                    &(S3ResponseHandler)
                    {
                        .propertiesCallback
                                          = globus_l_s3_rmdir_properties,
                        .completeCallback = globus_l_s3_rmdir_complete
                    },
                    rmdir_state);

            GlobusS3DebugOpStatus(S3_delete_object, rmdir_state);
        }
        while (S3_status_is_retryable(rmdir_state->status)
                && (rmdir_state->retries++) < globus_i_s3_max_retries);
    }
    else
    {
        rmdir_state->retries = 0;
        do
        {
            rmdir_state->status = S3StatusOK;
            rmdir_state->result = GLOBUS_SUCCESS;

            globus_i_s3_retry_delay(rmdir_state->retries);

            GlobusS3Info("REST: op=%p S3_delete_bucket(bucket=%s) try=%d\n",
                         (void *) rmdir_state->op,
                         rmdir_state->bucket_name,
                         rmdir_state->retries);
            S3_delete_bucket(
                    rmdir_state->bucket_context.protocol,
                    rmdir_state->bucket_context.uriStyle,
                    rmdir_state->bucket_context.accessKeyId,
                    rmdir_state->bucket_context.secretAccessKey,
                    rmdir_state->bucket_context.securityToken,
                    NULL,
                    rmdir_state->bucket_name,
                    NULL,
                    &(S3ResponseHandler)
                    {
                        .propertiesCallback
                                      = globus_l_s3_rmdir_properties,
                        .completeCallback
                                      = globus_l_s3_rmdir_complete
                    },
                    rmdir_state);

            GlobusS3DebugOpStatus(S3_delete_bucket, rmdir_state);
        }
        while (S3_status_is_retryable(rmdir_state->status)
                && (rmdir_state->retries++) < globus_i_s3_max_retries);
    }

    GlobusS3Info("finished_command: op=%p result=%#x\n",
                (void *) rmdir_state->op,
                (unsigned int) rmdir_state->result);
    globus_gridftp_server_finished_command(rmdir_state->op, rmdir_state->result, NULL);
    free(rmdir_state);

    return NULL;
}
/* globus_i_s3_rmdir() */

static
S3Status
globus_l_s3_rmdir_properties(
    const S3ResponseProperties         *properties,
    void                               *callback_arg)
{
    S3Status                            status = S3StatusOK;

    GlobusS3Enter();

    GlobusS3ExitStatus(status);

    return status;
}
/* globus_l_s3_rmdir_properties() */

static
void
globus_l_s3_rmdir_complete(
    S3Status                            status,
    const S3ErrorDetails               *error_details,
    void                               *callback_data)
{
    globus_l_s3_rmdir_t                *rmdir_state = callback_data;

    GlobusS3Enter();

    if (status != S3StatusOK && status != S3StatusHttpErrorNotFound)
    {
        rmdir_state->status = status;
        rmdir_state->result = GlobusGFSErrorS3(status, rmdir_state->pathname, error_details);
    }
    GlobusS3Exit();
}
/* globus_l_s3_rmdir_complete() */
// vim: ft=c:
