/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @file start_s3.c
 * Implementation of the s3 DSI start session
 */
#include "globus_i_s3.h"

static
globus_result_t
globus_l_evaluate_path(
    const char                         *path_format,
    char                              **pathp);

/**
 * @brief start
 * @details
 * This function is called when a new session is initialized, ie a user 
 * connects to the server.  This hook gives the dsi an opportunity to
 * reject the user, or set internal state that will be threaded through to all
 * other function calls associated with this session.
 */
void
globus_i_s3_start(
    globus_gfs_operation_t              op,
    globus_gfs_session_info_t          *session_info)
{
    globus_i_s3_handle_t               *s3_handle = NULL;
    char                                userline[512];
    char                               *accessKeyId = userline,
                                       *secretAccessKey = NULL,
                                       *securityToken = NULL;
    char                               *evaluated_path = NULL;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    if (globus_i_s3_map_filename != NULL)
    {
        FILE                           *mapfile = NULL;
        size_t                          namelen = strlen(session_info->username);
        result = globus_l_evaluate_path(
                globus_i_s3_map_filename,
                &evaluated_path);
        if (result != GLOBUS_SUCCESS)
        {
            goto bad_mapfile_path;
        }
        mapfile = fopen(evaluated_path, "r");
        if (mapfile == NULL)
        {
            result = GlobusGFSErrorSystemError("fopen", errno);

            goto open_path_fail;
        }

        while (!feof(mapfile))
        {
            const char                 *l = NULL;

            l = fgets(userline, sizeof(userline), mapfile);
            if (l == NULL)
            {
                result = GlobusGFSErrorGeneric("Unknown user");
                goto fail_parse;
            }
            else if (l[strlen(l)-1] != '\n')
            {
                // line too long
                result = GlobusGFSErrorMemory("user");
                goto fail_parse;
            }
            else if (l[0] == '#')
            {
                continue;
            }
            else if (strncmp(userline, session_info->username, namelen) == 0
                    && userline[namelen] == ';')
            {
                userline[strlen(userline)-1] = 0;
                memmove(userline,
                        userline + namelen + 1,
                        strlen(userline + namelen + 1) + 1);
                break;
            }
        }
fail_parse:
        fclose(mapfile);
        if (result != GLOBUS_SUCCESS)
        {
            goto bad_username;
        }
    }
    else
    {
        if (strlen(session_info->username) > (sizeof(userline) - 1))
        {
            result = GlobusGFSErrorGeneric("Invalid username");
            goto bad_username;
        }
        strcpy(userline, session_info->username);
    }

    if ((secretAccessKey = strchr(accessKeyId, ';')) != NULL)
    {
        *secretAccessKey++ = '\0';
    }
    if (accessKeyId == NULL || secretAccessKey == NULL)
    {
        result = GlobusGFSErrorParameter("username");
        goto bad_username;
    }
    if ((securityToken = strchr(secretAccessKey, ';')) != NULL)
    {
        *securityToken++ = '\0';
    }

    s3_handle = malloc(sizeof(globus_i_s3_handle_t));
    if (s3_handle == NULL)
    {
        result = GlobusGFSErrorMemory("s3_handle");
        goto malloc_fail;
    }
    *s3_handle = (globus_i_s3_handle_t)
    {
        .protocol                     = globus_i_s3_https
                                        ? S3ProtocolHTTPS
                                        : S3ProtocolHTTP,
        .uriStyle                     = globus_i_s3_use_virtual_host
                                        ? S3UriStyleVirtualHost
                                        : S3UriStylePath,
        .accessKeyId                  = strdup(accessKeyId),
        .secretAccessKey              = strdup(secretAccessKey),
        .securityToken                = securityToken ? strdup(securityToken) : NULL
    };

    if (s3_handle->accessKeyId == NULL)
    {
        result = GlobusGFSErrorMemory("accessKeyId");
        goto strdup_fail;
    }

    if (s3_handle->secretAccessKey == NULL)
    {
        result = GlobusGFSErrorMemory("secretAccessKey");
        goto strdup_fail;
    }

    if (result != GLOBUS_SUCCESS)
    {
strdup_fail:
        free(s3_handle->secretAccessKey);
        free(s3_handle->accessKeyId);
malloc_fail:
        free(s3_handle);
        s3_handle = NULL;
    }
bad_username:
open_path_fail:
    free(evaluated_path);
bad_mapfile_path:
    globus_gridftp_server_finished_session_start(
            op,
            result, 
            s3_handle,
            s3_handle ? s3_handle->accessKeyId : "ERROR",
            "/");
}
/* globus_l_s3_start() */

/**
 * @brief Replace special strings in a path
 * @details
 *    This function replaces some special substrings in the path passed as the
 *    path_format string path to a cache file and returns a copy of the
 *    string after replacement in *pathp.
 *
 *    The special strings that are replaced are:
 *    - $HOME
 *      The current user's home directory
 *    - $LOGNAME 
 *      The current user's login name
 *
 */
static
globus_result_t
globus_l_evaluate_path(
    /** [in] The original string to update */
    const char                         *path_format,
    /**
     * [out] A pointer to set to a copy of the path_format with special
     *       substrings replaced as described above. The caller is responsible
     *       for freeing this string.
     */
    char                              **pathp)
{
    char                               *buf = NULL;
    size_t                              bufsize = 0;
    char                               *path = NULL;
    globus_result_t                     result = GLOBUS_SUCCESS;
    uid_t                               id;
    struct passwd                       pwd, *res = NULL;
    int                                 rc = 0;
    const char                         *p = NULL;
    ptrdiff_t                           stretch = 0;
    ptrdiff_t                           logname_stretch = 0;
    ptrdiff_t                           home_stretch = 0;
    ptrdiff_t                           logname_len = 0;
    ptrdiff_t                           home_len = 0;
    const char                         *next_logname = NULL;
    const char                         *next_home = NULL;
    size_t                              i = 0;

    id = getuid();
    do
    {
        void *tmp;

        bufsize = bufsize ? bufsize * 2 : 64;
        tmp = realloc(buf, bufsize);

        if (!tmp)
        {
            result = GlobusGFSErrorMemory("pwd");

            goto buf_fail;
        }
        buf = tmp;
        rc = getpwuid_r(id, &pwd, buf, bufsize, &res);
    }
    while (rc != 0 && errno == ERANGE && bufsize < SIZE_MAX);

    if (rc != 0)
    {
        result = GlobusGFSErrorSystemError("getpwuid_r", errno);

        goto buf_fail;
    }

    if (pwd.pw_name != NULL)
    {
        logname_len = strlen(pwd.pw_name);
        logname_stretch = logname_len - strlen("$LOGNAME");
    }
    if (pwd.pw_dir != NULL)
    {
        home_len = strlen(pwd.pw_dir);
        home_stretch = home_len - strlen("$HOME");
    }
    p = path_format;
    while ((p = strstr(p, "$HOME")) != NULL)
    {
        p = p + 5;
        stretch += home_stretch;
    }
    p = path_format;
    while ((p = strstr(p, "$LOGNAME")) != NULL)
    {
        p = p + 8;
        stretch += logname_stretch;
    }
    path = malloc(strlen(path_format) + stretch + 1);
    if (path == NULL)
    {
        result = GlobusGFSErrorMemory("path");

        goto path_fail;
    }

    p = path_format;
    while (*p != 0)
    {
        size_t copy_len = strlen(p);
        next_logname = strstr(p, "$LOGNAME");
        next_home = strstr(p, "$HOME");

        if (next_home != NULL)
        {
            copy_len = next_home - p;
        }
        if (next_logname != NULL)
        {
            if ((next_logname - p) < copy_len)
            {
                copy_len = next_logname - p;
            }
        }
        memcpy(path + i, p, copy_len);
        i += copy_len;
        p += copy_len;

        if (p == next_logname)
        {
            if (pwd.pw_name != NULL)
            {
                memcpy(path + i, pwd.pw_name, logname_len);
                i += logname_len;
            }
            p += 8;
        }
        else if (p == next_home)
        {
            if (pwd.pw_dir != NULL)
            {
                memcpy(path + i, pwd.pw_dir, home_len);
                i += home_len;
            }
            p += 5;
        }
        assert (p != NULL);
    }
    path[i] = 0;


path_fail:
buf_fail:
    free(buf);

    *pathp = path;

    /**
     * @return
     *     If successful, this function returns `GLOBUS_SUCCESS`.
     *     If an error occurs, this function returns
     *     a globus_result_t referring to an error of one of the types 
     *     enumerated by #globus_google_drive_error_t.
     */
    
    return result;
}
/* globus_l_evaluate_path() */
// vim: ft=c:
