/*
 * Copyright 2016 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * @file stat.c
 * Implementation of the s3 DSI stat
 */

#include "globus_gridftp_server_s3_stat.h"

static
void *
globus_l_s3_stat_thread(
    void                               *user_arg);

/**
 * @brief stat
 * @details
 *
 * This interface function is called whenever the server needs 
 * information about a given file or resource.  It is called then an
 * LIST is sent by the client, when the server needs to verify that 
 * a file exists and has the proper permissions, etc.
 *
 * The s3 service provides a few different types of operations that
 * are useful for stat, depending on the stat options.
 */
void
globus_i_s3_stat(
    globus_gfs_operation_t              op,
    globus_gfs_stat_info_t             *stat_info,
    void                               *user_arg)
{
    globus_i_s3_handle_t               *s3_handle = user_arg;
    globus_i_s3_stat_t                 *stat_state = NULL;
    globus_thread_t                     t;
    int                                 rc;
    globus_result_t                     result = GLOBUS_SUCCESS;

    GlobusS3Enter();

    assert(stat_info != NULL);
    assert(s3_handle != NULL);

    GlobusS3Info("command=stat path=%s file_only=%s\n",
                 stat_info->pathname,
                 stat_info->file_only ? "true" : "false");


    stat_state = malloc(sizeof(globus_i_s3_stat_t) + strlen(stat_info->pathname) + 2);
    if (stat_state == NULL)
    {
        result = GlobusGFSErrorMemory("stat_state");

        goto fail;
    }

    *stat_state = (globus_i_s3_stat_t)
    {
        .op                           = op,
        .stat_info                    = stat_info,
    };

    strcpy(stat_state->pathname, stat_info->pathname);
    globus_i_s3_path_parse(stat_state->pathname, &stat_state->bucket_name, &stat_state->key);

    stat_state->bucket_context = GLOBUS_S3_BUCKET_CONTEXT(s3_handle, stat_state->bucket_name);

    if (strlen(stat_state->bucket_name) == 0)
    {
        if (stat_info->file_only)
        {
            stat_state->stat_type = GLOBUS_I_S3_STAT_ROOT;
        }
        else
        {
            stat_state->stat_type = GLOBUS_I_S3_LIST_BUCKETS;
        }
    }
    else if (stat_info->file_only)
    {
        if (stat_state->key == NULL)
        {
            stat_state->stat_type = GLOBUS_I_S3_STAT_BUCKET;
        }
        else
        {
            stat_state->stat_type = GLOBUS_I_S3_STAT_OBJECT;
        }
    }
    else
    {
        stat_state->stat_type = GLOBUS_I_S3_LIST_OBJECTS;
    }

    rc = globus_thread_create(
            &t,
            NULL,
            globus_l_s3_stat_thread,
            stat_state);


    if (rc != GLOBUS_SUCCESS)
    {
        result = GlobusGFSErrorGeneric("globus_thread_create");
fail:
        globus_gridftp_server_finished_stat( 
                op,
                result,
                NULL,
                0);

        free(stat_state);
    }

    GlobusS3Exit();

    return;
}
/* globus_i_s3_stat() */

static
void *
globus_l_s3_stat_thread(
    void                               *user_arg)
{

    globus_i_s3_stat_t                 *stat_state = user_arg;

    GlobusS3Enter();

    switch (stat_state->stat_type)
    {
        case GLOBUS_I_S3_STAT_ROOT:
            globus_i_s3_stat_root(stat_state);
            break;

        case GLOBUS_I_S3_LIST_BUCKETS:
            globus_i_s3_stat_list_buckets(stat_state);
            break;
        case GLOBUS_I_S3_STAT_BUCKET:
            globus_i_s3_stat_bucket(stat_state);
            break;

        case GLOBUS_I_S3_STAT_OBJECT:
            globus_i_s3_stat_object(stat_state);
            break;

        case GLOBUS_I_S3_LIST_OBJECTS:
            globus_i_s3_stat_list_objects(stat_state);
            break;
    }

    globus_gridftp_server_finished_stat(
            stat_state->op,
            stat_state->result,
            stat_state->stat,
            stat_state->stat_count);

    for (int i = 0; i < stat_state->stat_count; i++)
    {
        free(stat_state->stat[i].name);
    }
    free(stat_state);

    return NULL;
}
/* globus_l_s3_stat() */
// vim: ft=c:
